#include <stdio.h>
#include <stdarg.h>
#include "memalloc.h"
#include "list.h"
#include <debug.h>
#include <pthread.h>

static struct list free_list;   // list of free blocks

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* Comparator for Pintos list functions. */
static bool
mem_block_less(const struct list_elem *a, const struct list_elem *b, void *aux)
{
  struct free_block * ma = list_entry(a, struct free_block, elem);
  struct free_block * mb = list_entry(b, struct free_block, elem);
  return ma < mb;
}

/* Return number of elements in free list. */
size_t 
mem_sizeof_free_list(void)
{
  pthread_mutex_lock(&lock);
  size_t sz = list_size(&free_list);
  pthread_mutex_unlock(&lock);
  return sz;
}

/* Dump a list to screen. */
static void
mem_dump_list(struct list *l)
{
  struct list_elem *e;
  for (e = list_begin(l); e != list_end(l); e = list_next(e))
    {
      struct free_block *m = list_entry(e, struct free_block, elem);
      printf ("%p len=%d\n", m, (int) m->length);
    }
}

/* Dump free list to screen. */
void
mem_dump_free_list(void)
{
  printf("Dumping free block list:\n");
  mem_dump_list(&free_list);
}

/* Initialize allocator. */
void
mem_init(uint8_t *base, size_t length)
{
  list_init(&free_list);
  struct free_block * m = (struct free_block *)base;
  m->length = length;
  list_push_back(&free_list, &m->elem);
}

/* Allocate by finding free block using first-fit, in address order. */
void *
mem_alloc(size_t length)
{
  ASSERT (length % 4 == 0);

  /* must allocate minimum size to guarantee that freed block can
   * be inserted into freelist */
  length += sizeof (struct used_block);
  if (length < sizeof(struct free_block))
    length = sizeof(struct free_block);

  pthread_mutex_lock(&lock);
  struct list_elem *e;
  for (e = list_begin(&free_list); e != list_end(&free_list); e = list_next(e))
    {
      struct free_block *m = list_entry(e, struct free_block, elem);
      if (length <= m->length)
        {
          struct used_block *u = (struct used_block*)m;
          if (length + sizeof (*m) < m->length)
            {
              m->length -= length;
              u = (void *)m + m->length;        // bottom part
              u->length = length;
            }
          else
            list_remove(e);

          pthread_mutex_unlock(&lock);
          return u->data;
        }
    }

  pthread_mutex_unlock(&lock);
  return NULL;
}

/* Determine whether two blocks are adjacent. */
static bool
mem_block_is_adjacent(const struct free_block *left, 
                      const struct free_block *right)
{
  return (void *)left + left->length == (void *)right;
}

/* Coalesce two adjacent blocks. */
static void
mem_coalesce(struct free_block *left, struct free_block *right)
{
  list_remove(&right->elem);
  left->length += right->length;
}

/* Free memory, coalescing free list if necessary. */
void
mem_free(void *ptr)
{
  struct free_block *blk = ptr - offsetof(struct used_block, data);

  pthread_mutex_lock(&lock);

  /* Sort into free list. */
  list_insert_ordered(&free_list, &blk->elem, mem_block_less, NULL);

  /* Coalesce with right neighbor, if adjacent and not end of list. */
  struct list_elem *rightelem = list_next(&blk->elem);
  if (rightelem != list_end(&free_list))
    {
      struct free_block *right = list_entry(rightelem, struct free_block, elem);

      if (mem_block_is_adjacent(blk, right))
        mem_coalesce(blk, right);
    }

  /* Coalesce with left neighbor, if adjacent and not beginning of list. */
  struct list_elem *leftelem = list_prev(&blk->elem);
  if (leftelem != list_head(&free_list))
    {
      struct free_block *left = list_entry(leftelem, struct free_block, elem);

      if (mem_block_is_adjacent(left, blk))
        mem_coalesce(left, blk);
    }

  pthread_mutex_unlock(&lock);
}

// vim: sw=2
